/*----- PROTECTED REGION ID(DataVampire.cpp) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        DataVampire.cpp
//
// description : C++ source for the DataVampire class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               DataVampire are implemented in this file.
//
// project :     DataVampire
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include "DataVampire.h"
#include "DataVampireClass.h"

#include "acqthread.h"


/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	DataVampire.cpp

/**
 *  DataVampire class description:
 *    Get vectors of 1D or 2D data and calculate sum, mean, horpos, verpos, horsigma, versigma, numsamples
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name    |  Method name
//================================================================
//  State           |  Inherited (no method)
//  Status          |  Inherited (no method)
//  GetSum          |  get_sum
//  GetMean         |  get_mean
//  GetHorPos       |  get_hor_pos
//  GetVerPos       |  get_ver_pos
//  GetHorPosStd    |  get_hor_pos_std
//  GetVerPosStd    |  get_ver_pos_std
//  GetValidValues  |  get_valid_values
//  SetBackground   |  set_background
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  MinThreshold           |  Tango::DevDouble	Scalar
//  MaxThreshold           |  Tango::DevDouble	Scalar
//  CommandString          |  Tango::DevString	Scalar
//  PollingPeriod          |  Tango::DevLong	Scalar
//  PollingSamples         |  Tango::DevLong	Scalar
//  Sum                    |  Tango::DevDouble	Scalar
//  Mean                   |  Tango::DevDouble	Scalar
//  HorPos                 |  Tango::DevDouble	Scalar
//  VerPos                 |  Tango::DevDouble	Scalar
//  HorPosStd              |  Tango::DevDouble	Scalar
//  ValidValues            |  Tango::DevLong	Scalar
//  EnableAcquisition      |  Tango::DevBoolean	Scalar
//  CacheSize              |  Tango::DevLong	Scalar
//  AutoBackgroundSamples  |  Tango::DevLong	Scalar
//  PostAcqRoi             |  Tango::DevBoolean	Scalar
//  Width                  |  Tango::DevLong	Scalar
//  Height                 |  Tango::DevLong	Scalar
//  VerPosStd              |  Tango::DevDouble	Scalar
//  EnablePolling          |  Tango::DevBoolean	Scalar
//  RoiParam               |  Tango::DevLong	Spectrum  ( max = 4)
//  Spectrum               |  Tango::DevDouble	Spectrum  ( max = 1000000)
//  HorBackground          |  Tango::DevDouble	Spectrum  ( max = 1000000)
//  VerBackground          |  Tango::DevDouble	Spectrum  ( max = 1000000)
//  Image                  |  Tango::DevDouble	Image  ( max = 5000 x 5000)
//================================================================

namespace DataVampire_ns
{
/*----- PROTECTED REGION ID(DataVampire::namespace_starting) ENABLED START -----*/
/* clang-format on */
//	static initializations
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	DataVampire::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : DataVampire::DataVampire()
 *	Description : Constructors for a Tango device
 *                implementing the classDataVampire
 */
//--------------------------------------------------------
DataVampire::DataVampire(Tango::DeviceClass *cl, string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(DataVampire::constructor_1) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::constructor_1
}
//--------------------------------------------------------
DataVampire::DataVampire(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(DataVampire::constructor_2) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::constructor_2
}
//--------------------------------------------------------
DataVampire::DataVampire(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(DataVampire::constructor_3) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : DataVampire::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void DataVampire::delete_device()
{
	DEBUG_STREAM << "DataVampire::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::delete_device) ENABLED START -----*/
	/* clang-format on */
	//	Delete device allocated objects
	/* clang-format off */
	abort_flag = true;
	if (acqloop) {
		acqloop->join(NULL);
		acqloop = 0;
	}	
	
	if (attr_CommandString_read) {
		delete [] *attr_CommandString_read;
		attr_CommandString_read = 0;
	}

	if (device)
		delete device;

	if (bn_device)
		delete bn_device;

	delete mutex;

	free_cache_buffer();

	/*----- PROTECTED REGION END -----*/	//	DataVampire::delete_device
	delete[] attr_MinThreshold_read;
	delete[] attr_MaxThreshold_read;
	delete[] attr_CommandString_read;
	delete[] attr_PollingPeriod_read;
	delete[] attr_PollingSamples_read;
	delete[] attr_Sum_read;
	delete[] attr_Mean_read;
	delete[] attr_HorPos_read;
	delete[] attr_VerPos_read;
	delete[] attr_HorPosStd_read;
	delete[] attr_ValidValues_read;
	delete[] attr_EnableAcquisition_read;
	delete[] attr_CacheSize_read;
	delete[] attr_AutoBackgroundSamples_read;
	delete[] attr_PostAcqRoi_read;
	delete[] attr_Width_read;
	delete[] attr_Height_read;
	delete[] attr_VerPosStd_read;
	delete[] attr_EnablePolling_read;
	delete[] attr_RoiParam_read;
	delete[] attr_Spectrum_read;
	delete[] attr_HorBackground_read;
	delete[] attr_VerBackground_read;
	delete[] attr_Image_read;
}

//--------------------------------------------------------
/**
 *	Method      : DataVampire::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void DataVampire::init_device()
{
	DEBUG_STREAM << "DataVampire::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::init_device_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialization before get_device_property() call
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::init_device_before
	

	//	Get the device properties from database
	get_device_property();
	
	attr_MinThreshold_read = new Tango::DevDouble[1];
	attr_MaxThreshold_read = new Tango::DevDouble[1];
	attr_CommandString_read = new Tango::DevString[1];
	attr_PollingPeriod_read = new Tango::DevLong[1];
	attr_PollingSamples_read = new Tango::DevLong[1];
	attr_Sum_read = new Tango::DevDouble[1];
	attr_Mean_read = new Tango::DevDouble[1];
	attr_HorPos_read = new Tango::DevDouble[1];
	attr_VerPos_read = new Tango::DevDouble[1];
	attr_HorPosStd_read = new Tango::DevDouble[1];
	attr_ValidValues_read = new Tango::DevLong[1];
	attr_EnableAcquisition_read = new Tango::DevBoolean[1];
	attr_CacheSize_read = new Tango::DevLong[1];
	attr_AutoBackgroundSamples_read = new Tango::DevLong[1];
	attr_PostAcqRoi_read = new Tango::DevBoolean[1];
	attr_Width_read = new Tango::DevLong[1];
	attr_Height_read = new Tango::DevLong[1];
	attr_VerPosStd_read = new Tango::DevDouble[1];
	attr_EnablePolling_read = new Tango::DevBoolean[1];
	attr_RoiParam_read = new Tango::DevLong[4];
	attr_Spectrum_read = new Tango::DevDouble[1000000];
	attr_HorBackground_read = new Tango::DevDouble[1000000];
	attr_VerBackground_read = new Tango::DevDouble[1000000];
	attr_Image_read = new Tango::DevDouble[5000*5000];
	/*----- PROTECTED REGION ID(DataVampire::init_device) ENABLED START -----*/
	/* clang-format on */
	//	Initialize device
	/* clang-format off */

	*attr_CommandString_read = new char[DATAVAMPIRE_MAX_STR_LEN];

	*attr_AutoBackgroundSamples_read = 0;
	*attr_MinThreshold_read = -99999;
	*attr_MaxThreshold_read = 99999;
	*attr_CacheSize_read = 100;
	*attr_PostAcqRoi_read = false;
	*attr_EnableAcquisition_read = true;

	num_rows_image = 0;
	num_cols_image = 0;
	num_cols_spectrum = 0;

	roi_dim = 0;
	for (int i = 0; i < 4; i++) {
		attr_RoiParam_read[i] = roiParam[i];
		if (attr_RoiParam_read[i] >= 0)
			roi_dim++;
	}

	std::vector<std::string> results;
	std::string new_configuration(commandString);
	string_explode(new_configuration, "/", &results);

	command_name = results[results.size()-1];
	size_t pos = new_configuration.find(command_name);
	if (pos != std::string::npos) { // if str2 found in str1
        	new_configuration.erase(pos-1, command_name.length()+1); // erase str2 from str1
    	}	
	device_name = new_configuration;

	try {
		device = new Tango::DeviceProxy(device_name);

		Tango::CommandInfo cmd_info = device->command_query(command_name);
		cmd_in_type = cmd_info.in_type;
		cmd_out_type = cmd_info.out_type;
		snprintf(*attr_CommandString_read, DATAVAMPIRE_MAX_STR_LEN, "%s", new_configuration.c_str());
		set_state(Tango::ON);
		set_status("Connected to device");
	}	
	catch(...) {
		set_state(Tango::FAULT);
		set_status("Failed to execute device command, check CommandString property and restart device");
		printf("Error checking %s\n",new_configuration.c_str());
	}	

	/*
	std::vector<std::string> results_a;
	std::string bn_attr(bunchNumberString);
	string_explode(bn_attr, "/", &results_a);	
	bn_attr_name = results[results_a.size()-1];
	pos = bn_attr.find(bn_attr_name);
	if (pos != std::string::npos) { // if str2 found in str1
        	bn_attr.erase(pos-1, bn_attr_name.length()+1); // erase str2 from str1
    	}	
	bn_device_name = bn_attr;

	printf("5\n");
	
	try {
		bn_device = new Tango::DeviceProxy(bn_device_name);
	}	
	catch(...) {
		set_state(Tango::FAULT);
		set_status("Failed to connect to BunchNumber device");	
	}
	*/

	

		
		
	cache_buffer = 0;
	allocate_cache_buffer();

	/*
	mutex = new omni_mutex();
	
	acqloop = 0;
	acqloop = new acqthread(this);
	acqloop->start_undetached();
	*/

	/*----- PROTECTED REGION END -----*/	//	DataVampire::init_device
}

//--------------------------------------------------------
/**
 *	Method      : DataVampire::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void DataVampire::get_device_property()
{
	/*----- PROTECTED REGION ID(DataVampire::get_device_property_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialize property data members
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("HorCalib"));
	dev_prop.push_back(Tango::DbDatum("VerCalib"));
	dev_prop.push_back(Tango::DbDatum("RoiParam"));
	dev_prop.push_back(Tango::DbDatum("CommandString"));
	dev_prop.push_back(Tango::DbDatum("BunchNumberString"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on DataVampireClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		DataVampireClass	*ds_class =
			(static_cast<DataVampireClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize HorCalib from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  horCalib;
		else {
			//	Try to initialize HorCalib from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  horCalib;
		}
		//	And try to extract HorCalib value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  horCalib;

		//	Try to initialize VerCalib from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  verCalib;
		else {
			//	Try to initialize VerCalib from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  verCalib;
		}
		//	And try to extract VerCalib value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  verCalib;

		//	Try to initialize RoiParam from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  roiParam;
		else {
			//	Try to initialize RoiParam from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  roiParam;
		}
		//	And try to extract RoiParam value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  roiParam;

		//	Try to initialize CommandString from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  commandString;
		else {
			//	Try to initialize CommandString from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  commandString;
		}
		//	And try to extract CommandString value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  commandString;

		//	Try to initialize BunchNumberString from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  bunchNumberString;
		else {
			//	Try to initialize BunchNumberString from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  bunchNumberString;
		}
		//	And try to extract BunchNumberString value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  bunchNumberString;

	}

	/*----- PROTECTED REGION ID(DataVampire::get_device_property_after) ENABLED START -----*/
	/* clang-format on */
	//	Check device property data members init
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_device_property_after
}

//--------------------------------------------------------
/**
 *	Method      : DataVampire::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void DataVampire::always_executed_hook()
{
	DEBUG_STREAM << "DataVampire::always_executed_hook()  " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::always_executed_hook) ENABLED START -----*/
	/* clang-format on */
	//	code always executed before all requests
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : DataVampire::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void DataVampire::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "DataVampire::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method      : DataVampire::write_attr_hardware()
 *	Description : Hardware writing for attributes
 */
//--------------------------------------------------------
void DataVampire::write_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "DataVampire::write_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::write_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute MinThreshold related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_MinThreshold(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_MinThreshold(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_MinThreshold) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_MinThreshold_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_MinThreshold
}
//--------------------------------------------------------
/**
 *	Write attribute MinThreshold related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_MinThreshold(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_MinThreshold(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_MinThreshold) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_MinThreshold_read = w_val;
	
	
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_MinThreshold
}
//--------------------------------------------------------
/**
 *	Read attribute MaxThreshold related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_MaxThreshold(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_MaxThreshold(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_MaxThreshold) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_MaxThreshold_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_MaxThreshold
}
//--------------------------------------------------------
/**
 *	Write attribute MaxThreshold related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_MaxThreshold(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_MaxThreshold(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_MaxThreshold) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_MaxThreshold_read = w_val;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_MaxThreshold
}
//--------------------------------------------------------
/**
 *	Read attribute CommandString related method
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_CommandString(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_CommandString(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_CommandString) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_CommandString_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_CommandString
}
//--------------------------------------------------------
/**
 *	Write attribute CommandString related method
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_CommandString(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_CommandString(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevString	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_CommandString) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	std::vector<std::string> results;
	std::string new_configuration(w_val);
	string_explode(new_configuration, "/", &results);

	command_name = results[results.size()-1];
	size_t pos = new_configuration.find(command_name);
	if (pos != std::string::npos) { // if str2 found in str1
        	new_configuration.erase(pos-1, command_name.length()+1); // erase str2 from str1
    	}	
	device_name = new_configuration;

	if (device)
		delete device;
	device = new Tango::DeviceProxy(device_name);

	Tango::CommandInfo cmd_info = device->command_query(command_name);
	cmd_in_type = cmd_info.in_type;
	cmd_out_type = cmd_info.out_type;

	snprintf(*attr_CommandString_read, DATAVAMPIRE_MAX_STR_LEN, "%s", new_configuration.c_str());
	store_command_string(w_val);
	
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_CommandString
}
//--------------------------------------------------------
/**
 *	Read attribute PollingPeriod related method
 *	Description: -1: disabled
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_PollingPeriod(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_PollingPeriod(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_PollingPeriod) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_PollingPeriod_read);
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_PollingPeriod
}
//--------------------------------------------------------
/**
 *	Write attribute PollingPeriod related method
 *	Description: -1: disabled
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_PollingPeriod(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_PollingPeriod(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_PollingPeriod) ENABLED START -----*/
	
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_PollingPeriod
}
//--------------------------------------------------------
/**
 *	Read attribute PollingSamples related method
 *	Description: -1: disabled
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_PollingSamples(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_PollingSamples(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_PollingSamples) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_PollingSamples_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_PollingSamples
}
//--------------------------------------------------------
/**
 *	Write attribute PollingSamples related method
 *	Description: -1: disabled
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_PollingSamples(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_PollingSamples(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_PollingSamples) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	*attr_PollingSamples_read = w_val;
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_PollingSamples
}
//--------------------------------------------------------
/**
 *	Read attribute Sum related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_Sum(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_Sum(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_Sum) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_Sum_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_Sum
}
//--------------------------------------------------------
/**
 *	Read attribute Mean related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_Mean(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_Mean(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_Mean) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_Mean_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_Mean
}
//--------------------------------------------------------
/**
 *	Read attribute HorPos related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_HorPos(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_HorPos(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_HorPos) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_HorPos_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_HorPos
}
//--------------------------------------------------------
/**
 *	Read attribute VerPos related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_VerPos(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_VerPos(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_VerPos) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_VerPos_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_VerPos
}
//--------------------------------------------------------
/**
 *	Read attribute HorPosStd related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_HorPosStd(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_HorPosStd(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_HorPosStd) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_HorPosStd_read);
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_HorPosStd
}
//--------------------------------------------------------
/**
 *	Read attribute ValidValues related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_ValidValues(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_ValidValues(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_ValidValues) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_ValidValues_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_ValidValues
}
//--------------------------------------------------------
/**
 *	Read attribute EnableAcquisition related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_EnableAcquisition(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_EnableAcquisition(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_EnableAcquisition) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_EnableAcquisition_read);
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_EnableAcquisition
}
//--------------------------------------------------------
/**
 *	Write attribute EnableAcquisition related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_EnableAcquisition(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_EnableAcquisition(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_EnableAcquisition) ENABLED START -----*/
	*attr_EnableAcquisition_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_EnableAcquisition
}
//--------------------------------------------------------
/**
 *	Read attribute CacheSize related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_CacheSize(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_CacheSize(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_CacheSize) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_CacheSize_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_CacheSize
}
//--------------------------------------------------------
/**
 *	Write attribute CacheSize related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_CacheSize(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_CacheSize(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_CacheSize) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	*attr_CacheSize_read = w_val;

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_CacheSize
}
//--------------------------------------------------------
/**
 *	Read attribute AutoBackgroundSamples related method
 *	Description: 0 to disable
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_AutoBackgroundSamples(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_AutoBackgroundSamples(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_AutoBackgroundSamples) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_AutoBackgroundSamples_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_AutoBackgroundSamples
}
//--------------------------------------------------------
/**
 *	Write attribute AutoBackgroundSamples related method
 *	Description: 0 to disable
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_AutoBackgroundSamples(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_AutoBackgroundSamples(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_AutoBackgroundSamples) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	*attr_AutoBackgroundSamples_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_AutoBackgroundSamples
}
//--------------------------------------------------------
/**
 *	Read attribute PostAcqRoi related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_PostAcqRoi(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_PostAcqRoi(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_PostAcqRoi) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_PostAcqRoi_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_PostAcqRoi
}
//--------------------------------------------------------
/**
 *	Write attribute PostAcqRoi related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_PostAcqRoi(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_PostAcqRoi(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_PostAcqRoi) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	*attr_PostAcqRoi_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_PostAcqRoi
}
//--------------------------------------------------------
/**
 *	Read attribute Width related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_Width(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_Width(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_Width) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Width_read);
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_Width
}
//--------------------------------------------------------
/**
 *	Write attribute Width related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_Width(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_Width(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_Width) ENABLED START -----*/
	*attr_Width_read = w_val;	
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_Width
}
//--------------------------------------------------------
/**
 *	Read attribute Height related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_Height(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_Height(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_Height) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Height_read);
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_Height
}
//--------------------------------------------------------
/**
 *	Write attribute Height related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_Height(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_Height(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_Height) ENABLED START -----*/
	*attr_Height_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_Height
}
//--------------------------------------------------------
/**
 *	Read attribute VerPosStd related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_VerPosStd(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_VerPosStd(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_VerPosStd) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_VerPosStd_read);
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_VerPosStd
}
//--------------------------------------------------------
/**
 *	Read attribute EnablePolling related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::read_EnablePolling(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_EnablePolling(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_EnablePolling) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_EnablePolling_read);
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_EnablePolling
}
//--------------------------------------------------------
/**
 *	Write attribute EnablePolling related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void DataVampire::write_EnablePolling(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_EnablePolling(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_EnablePolling) ENABLED START -----*/
	
	
	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_EnablePolling
}
//--------------------------------------------------------
/**
 *	Read attribute RoiParam related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void DataVampire::read_RoiParam(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_RoiParam(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_RoiParam) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_RoiParam_read, roi_dim);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_RoiParam
}
//--------------------------------------------------------
/**
 *	Write attribute RoiParam related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void DataVampire::write_RoiParam(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "DataVampire::write_RoiParam(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevLong	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(DataVampire::write_RoiParam) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	store_roi_param(w_val, w_length);

	/*----- PROTECTED REGION END -----*/	//	DataVampire::write_RoiParam
}
//--------------------------------------------------------
/**
 *	Read attribute Spectrum related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 1000000
 */
//--------------------------------------------------------
void DataVampire::read_Spectrum(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_Spectrum(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_Spectrum) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_Spectrum_read, num_cols_spectrum);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_Spectrum
}
//--------------------------------------------------------
/**
 *	Read attribute HorBackground related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 1000000
 */
//--------------------------------------------------------
void DataVampire::read_HorBackground(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_HorBackground(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_HorBackground) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	int hor_size = 0;
	if (roi_dim == 2)
		hor_size = attr_RoiParam_read[1] - attr_RoiParam_read[0] + 1;
	else if (roi_dim == 4)
		hor_size = attr_RoiParam_read[2] - attr_RoiParam_read[0] + 1;

	attr.set_value(attr_HorBackground_read, hor_size);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_HorBackground
}
//--------------------------------------------------------
/**
 *	Read attribute VerBackground related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 1000000
 */
//--------------------------------------------------------
void DataVampire::read_VerBackground(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_VerBackground(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_VerBackground) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	int ver_size = 0;
	if (roi_dim == 2)
		ver_size = 0;
	else if (roi_dim == 4)
		ver_size = attr_RoiParam_read[3] - attr_RoiParam_read[1] + 1;

	attr.set_value(attr_VerBackground_read, ver_size);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_VerBackground
}
//--------------------------------------------------------
/**
 *	Read attribute Image related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Image max = 5000 x 5000
 */
//--------------------------------------------------------
void DataVampire::read_Image(Tango::Attribute &attr)
{
	DEBUG_STREAM << "DataVampire::read_Image(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(DataVampire::read_Image) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_Image_read, num_cols_image, num_rows_image);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::read_Image
}

//--------------------------------------------------------
/**
 *	Method      : DataVampire::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void DataVampire::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(DataVampire::add_dynamic_attributes) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic attributes if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command GetSum related method
 *	Description: 
 *
 *	@param argin mode,samples; mode,bunch start,bunch end
 *	@returns values
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *DataVampire::get_sum(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "DataVampire::GetSum()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::get_sum) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	acq_buffer_t *rec_data;
	recover_data(argin, &rec_data);
		

	argout = new Tango::DevVarDoubleArray();
	argout->length(rec_data->num_samples);

	for (int i = 0; i < rec_data->num_samples; i++)
		(*argout)[i] = rec_data->sum[i];

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_sum
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetMean related method
 *	Description: 
 *
 *	@param argin mode,samples; mode,bunch start,bunch end
 *	@returns values
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *DataVampire::get_mean(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "DataVampire::GetMean()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::get_mean) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	acq_buffer_t *rec_data;
	recover_data(argin, &rec_data);
		

	argout = new Tango::DevVarDoubleArray();
	argout->length(rec_data->num_samples);

	for (int i = 0; i < rec_data->num_samples; i++)
		(*argout)[i] = rec_data->mean[i];

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_mean
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetHorPos related method
 *	Description: 
 *
 *	@param argin mode,samples; mode,bunch start,bunch end
 *	@returns values
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *DataVampire::get_hor_pos(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "DataVampire::GetHorPos()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::get_hor_pos) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	acq_buffer_t *rec_data;
	recover_data(argin, &rec_data);
		

	argout = new Tango::DevVarDoubleArray();
	argout->length(rec_data->num_samples);

	for (int i = 0; i < rec_data->num_samples; i++)
		(*argout)[i] = rec_data->horpos[i];

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_hor_pos
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetVerPos related method
 *	Description: 
 *
 *	@param argin mode,samples; mode,bunch start,bunch end
 *	@returns values
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *DataVampire::get_ver_pos(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "DataVampire::GetVerPos()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::get_ver_pos) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	acq_buffer_t *rec_data;
	recover_data(argin, &rec_data);
		

	argout = new Tango::DevVarDoubleArray();
	argout->length(rec_data->num_samples);

	for (int i = 0; i < rec_data->num_samples; i++)
		(*argout)[i] = rec_data->verpos[i];

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_ver_pos
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetHorPosStd related method
 *	Description: 
 *
 *	@param argin mode,samples; mode,bunch start,bunch end
 *	@returns values
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *DataVampire::get_hor_pos_std(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "DataVampire::GetHorPosStd()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::get_hor_pos_std) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	acq_buffer_t *rec_data;
	recover_data(argin, &rec_data);
		

	argout = new Tango::DevVarDoubleArray();
	argout->length(rec_data->num_samples);

	for (int i = 0; i < rec_data->num_samples; i++)
		(*argout)[i] = rec_data->horposstd[i];

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_hor_pos_std
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetVerPosStd related method
 *	Description: 
 *
 *	@param argin mode,samples; mode,bunch start,bunch end
 *	@returns values
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *DataVampire::get_ver_pos_std(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "DataVampire::GetVerPosStd()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::get_ver_pos_std) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	acq_buffer_t *rec_data;
	recover_data(argin, &rec_data);
		

	argout = new Tango::DevVarDoubleArray();
	argout->length(rec_data->num_samples);

	for (int i = 0; i < rec_data->num_samples; i++)
		(*argout)[i] = rec_data->verposstd[i];

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_ver_pos_std
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetValidValues related method
 *	Description: 
 *
 *	@param argin mode,samples; mode,bunch start,bunch end
 *	@returns values
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *DataVampire::get_valid_values(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "DataVampire::GetValidValues()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::get_valid_values) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	acq_buffer_t *rec_data;
	recover_data(argin, &rec_data);
		

	argout = new Tango::DevVarDoubleArray();
	argout->length(rec_data->num_samples);

	for (int i = 0; i < rec_data->num_samples; i++)
		(*argout)[i] = rec_data->validvalues[i];
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::get_valid_values
	return argout;
}
//--------------------------------------------------------
/**
 *	Command SetBackground related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void DataVampire::set_background()
{
	DEBUG_STREAM << "DataVampire::SetBackground()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(DataVampire::set_background) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::set_background
}
//--------------------------------------------------------
/**
 *	Method      : DataVampire::add_dynamic_commands()
 *	Description : Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void DataVampire::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(DataVampire::add_dynamic_commands) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic commands if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	DataVampire::add_dynamic_commands
}

/*----- PROTECTED REGION ID(DataVampire::namespace_ending) ENABLED START -----*/
/* clang-format on */
//	Additional Methods
/* clang-format off */





void DataVampire::cmd_extract_values(Tango::DeviceData *cmd_value, std::vector<double> &val, std::vector<std::string> &val_string)
{

	Tango::DevState stval;
	std::vector<Tango::DevState> v_st;
	std::vector<Tango::DevULong> v_ulo;
	std::vector<Tango::DevUChar> v_uch;
	std::vector<Tango::DevShort> v_sh;
	std::vector<Tango::DevUShort> v_ush;
	std::vector<Tango::DevLong> v_lo;
	std::vector<Tango::DevDouble> v_do;
	std::vector<Tango::DevFloat> v_fl;
	std::vector<Tango::DevBoolean> v_bo;
	std::vector<Tango::DevLong64> v_lo64;
	std::vector<Tango::DevULong64> v_ulo64;
	//vector<string> v_string;

	long cmd_type = cmd_value->get_type();

	val.clear();

	if ((cmd_type == Tango::DEV_UCHAR) || (cmd_type == Tango::DEVVAR_CHARARRAY)) 
	{
		*cmd_value >> v_uch;
		for(std::vector<Tango::DevUChar>::iterator it = v_uch.begin(); it != v_uch.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_SHORT) || (cmd_type == Tango::DEVVAR_SHORTARRAY)) 
	{
		*cmd_value >> v_sh;
		for(std::vector<Tango::DevShort>::iterator  it = v_sh.begin(); it != v_sh.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_USHORT) || (cmd_type == Tango::DEVVAR_USHORTARRAY)) 
	{
		*cmd_value >> v_ush;
		for(std::vector<Tango::DevUShort>::iterator  it = v_ush.begin(); it != v_ush.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_LONG) || (cmd_type == Tango::DEVVAR_LONGARRAY)) 
	{
		*cmd_value >> v_lo;
		for(std::vector<Tango::DevLong>::iterator  it = v_lo.begin(); it != v_lo.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_LONG64) || (cmd_type == Tango::DEVVAR_LONG64ARRAY)) 
	{
		*cmd_value >> v_lo64;
		for(std::vector<Tango::DevLong64>::iterator  it = v_lo64.begin(); it != v_lo64.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if (cmd_type == Tango::DEV_STATE)
	{
		//*cmd_value >> v_st;
		//TODO: why extract_read not working with State
		std::cout << "GG: Do not compile cmd_value->get_type() == Tango::DEV_STATE ???" << std::endl;
		if(v_st.size() == 0)
		{
			*cmd_value >> stval;
			v_st.push_back(stval);
		}
		for(std::vector<Tango::DevState>::iterator it = v_st.begin(); it != v_st.end(); it++)
		{
			val.push_back((double)(*it));		//convert all to double
		}
	}
	else if ((cmd_type == Tango::DEV_ULONG) || (cmd_type == Tango::DEVVAR_ULONGARRAY))
	{
		*cmd_value >> v_ulo;
		for(std::vector<Tango::DevULong>::iterator  it = v_ulo.begin(); it != v_ulo.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_ULONG64) || (cmd_type == Tango::DEVVAR_ULONG64ARRAY)) 
	{
		*cmd_value >> v_ulo64;
		for(std::vector<Tango::DevULong64>::iterator  it = v_ulo64.begin(); it != v_ulo64.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_DOUBLE) || (cmd_type == Tango::DEVVAR_DOUBLEARRAY)) 
	{
		*cmd_value >> v_do;

		for(std::vector<Tango::DevDouble>::iterator  it = v_do.begin(); it != v_do.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_FLOAT) || (cmd_type == Tango::DEVVAR_FLOATARRAY)) 
	{
		*cmd_value >> v_fl;
		for(std::vector<Tango::DevFloat>::iterator  it = v_fl.begin(); it != v_fl.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_BOOLEAN) || (cmd_type == Tango::DEVVAR_BOOLEANARRAY)) 
	{
		//*cmd_value >> v_bo; // GG do not compile ????
		std::cout << "GG: Do not compile cmd_value->get_type() == Tango::DEV_BOOLEAN ???" << std::endl;
		for(std::vector<Tango::DevBoolean>::iterator  it = v_bo.begin(); it != v_bo.end(); it++)
			val.push_back((double)(*it));		//convert all to double
	}
	else if ((cmd_type == Tango::DEV_STRING) || (cmd_type == Tango::DEVVAR_STRINGARRAY)) 
	{
		/*attr_value->extract_read(v_string);
		val_string = *(v_string.begin());*/
		*cmd_value >> val_string;
	}
	else
	{
		std::ostringstream o;
		o << "unknown type" << std::ends;
		throw o.str();
	}

}


Tango::DeviceData DataVampire::cmd_insert_values(long cmd_type, std::vector<double> val, std::vector<std::string> val_string)
{
	Tango::DevState st;
	std::vector<Tango::DevState> v_st;
	std::vector<Tango::DevULong> v_ulo;
	Tango::DevULong ulo;
	std::vector<Tango::DevULong64> v_ulo64;
	Tango::DevULong64 ulo64;
	std::vector<Tango::DevUChar> v_uch;
	Tango::DevUChar uch;
	std::vector<Tango::DevShort> v_sh;
	Tango::DevShort sh;
	std::vector<Tango::DevUShort> v_ush;
	Tango::DevUShort ush;
	std::vector<Tango::DevLong> v_lo;
	Tango::DevLong lo;
	std::vector<Tango::DevLong64> v_lo64;
	Tango::DevLong64 lo64;
	std::vector<Tango::DevDouble> v_do;
	Tango::DevDouble _do;
	std::vector<Tango::DevFloat> v_fl;
	Tango::DevFloat fl;
	std::vector<Tango::DevBoolean> v_bo;
	Tango::DevBoolean bo;
	std::vector<std::string> v_string;
	std::string _string;
	Tango::DeviceData din_value;
	
	if ((cmd_type == Tango::DEV_UCHAR) || (cmd_type == Tango::DEVVAR_CHARARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_uch.push_back((Tango::DevUChar)(*it));		//convert all from double
			//din_value << v_uch;
			std::cout << "GG: Do not compile cmd_type == Tango::DEV_UCHAR ???" << std::endl;
		}
		else
		{
			uch = (Tango::DevUChar)(*val.begin());
			// din_value << uch;
			std::cout << "GG: Do not compile cmd_type == Tango::DEV_UCHAR ???" << std::endl;
		}
	}
	else if ((cmd_type == Tango::DEV_SHORT) || (cmd_type == Tango::DEVVAR_SHORTARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_sh.push_back((Tango::DevShort)(*it));		//convert all from double
			din_value << v_sh;
		}
		else
		{
			sh = (Tango::DevShort)(*val.begin());
			din_value << sh;
		}
	}
	else if ((cmd_type == Tango::DEV_USHORT) || (cmd_type == Tango::DEVVAR_USHORTARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_ush.push_back((Tango::DevUShort)(*it));		//convert all from double
			din_value << v_ush;
		}
		else
		{
			ush = (Tango::DevUShort)(*val.begin());
			din_value << ush;
		}
	}
	else if ((cmd_type == Tango::DEV_LONG) || (cmd_type == Tango::DEVVAR_LONGARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_lo.push_back((Tango::DevLong)(*it));		//convert all from double
			din_value << v_lo;
		}
		else
		{
			lo = (Tango::DevLong)(*val.begin());
			din_value << lo;
		}
	}
	else if ((cmd_type == Tango::DEV_LONG64) || (cmd_type == Tango::DEVVAR_LONG64ARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_lo64.push_back((Tango::DevLong64)(*it));		//convert all from double
			din_value << v_lo64;
		}
		else
		{
			lo64 = (Tango::DevLong64)(*val.begin());
			din_value << lo64;
		}
	}
	else if (cmd_type == Tango::DEV_STATE)
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_st.push_back((Tango::DevState)(*it));		//convert all from double
			//din_value << v_st;
			std::cout << "GG: Do not compile cmd_type == Tango::DEV_STATE ???" << std::endl;
		}
		else
		{
			st = (Tango::DevState)(*(val.begin()));
			din_value << st;
		}
	}
	else if ((cmd_type == Tango::DEV_ULONG) || (cmd_type == Tango::DEVVAR_ULONGARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_ulo.push_back((Tango::DevULong)(*it));		//convert all from double
			din_value << v_ulo;
		}
		else
		{
			ulo = (Tango::DevULong)(*(val.begin()));
			din_value << ulo;
		}
	}
	else if ((cmd_type == Tango::DEV_ULONG64) || (cmd_type == Tango::DEVVAR_ULONG64ARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_ulo64.push_back((Tango::DevULong64)(*it));		//convert all from double
			din_value << v_ulo64;
		}
		else
		{
			ulo64 = (Tango::DevULong64)(*(val.begin()));
			din_value << ulo64;
		}
	}
	else if ((cmd_type == Tango::DEV_DOUBLE) || (cmd_type == Tango::DEVVAR_DOUBLEARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
			{
				v_do.push_back((Tango::DevDouble)(*it));		//convert all from double
			}
			din_value << v_do;
		}
		else
		{
			_do = (Tango::DevDouble)(*(val.begin()));
			din_value << _do;
		}
	}
	else if ((cmd_type == Tango::DEV_FLOAT) || (cmd_type == Tango::DEVVAR_FLOATARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_fl.push_back((Tango::DevFloat)(*it));		//convert all from double
			din_value << v_fl;
		}
		else
		{
			fl = (Tango::DevFloat)(*(val.begin()));
			din_value << fl;
		}
	}
	else if ((cmd_type == Tango::DEV_BOOLEAN) || (cmd_type == Tango::DEVVAR_BOOLEANARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<double>::iterator it = val.begin(); it != val.end(); it++)
				v_bo.push_back((Tango::DevBoolean)(*it));		//convert all from double
			// din_value << v_bo;
				std::cout << "GG: Do not compile cmd_type == Tango::DEV_BOOLEAN ???" << std::endl;
		}
		else
		{
			bo = (Tango::DevBoolean)(*(val.begin()));
			din_value << bo;
		}
	}
	else if ((cmd_type == Tango::DEV_STRING) || (cmd_type == Tango::DEVVAR_STRINGARRAY)) 
	{
		if(val.size() > 1)
		{
			for(std::vector<std::string>::iterator it = val_string.begin(); it != val_string.end(); it++)
				v_string.push_back(*it);		//convert all from string
			din_value << v_string;
		}
		else
		{
			_string = *(val_string.begin());
			din_value << _string;
		}
	}
	else
	{
		std::ostringstream o;
		o << "unknown type" << std::ends;
		throw o.str();
	}

	return din_value;
}


//+------------------------------------------------------------------
/**
 *	method:	DataVampire::string_explode
 * Explode a string 
*/
//+------------------------------------------------------------------
void DataVampire::string_explode(std::string str, std::string separator, std::vector<std::string>* results)
{
	size_t found;
	
	found = str.find_first_of(separator);
	while(found != std::string::npos) {
		if(found > 0) {
			results->push_back(str.substr(0,found));
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		results->push_back(str);
	}

}



//+------------------------------------------------------------------
/**
 * method:	DataVampire::store_roi_param
 * Store command string 
*/
//+------------------------------------------------------------------
void DataVampire::store_roi_param(const Tango::DevLong *w_val, int size)
{

	Tango::Database *db = new Tango::Database();
	char conf_str[20];

	/* manual ROI x1,y1,x2,y2 */
	try {
		Tango::DbDatum config("RoiParam");
		Tango::DbData db_data_put;
		std::vector<std::string> val;
		for (int i = 0; i < size; i++) {
			memset(conf_str,0,20);
			sprintf(conf_str,"%d", w_val[i]);
			val.push_back(conf_str);
		}
		roi_dim = size;
		for (int i = size; i < 4; i++) {
			memset(conf_str,0,20);
			sprintf(conf_str,"-1");
			val.push_back(conf_str);
		}

		config << val;
		db_data_put.push_back(config);
		db->put_device_property(get_name(), db_data_put);
	}
	catch (Tango::DevFailed &e) {
		delete db;
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			__FUNCTION__);				
	}
	delete db;

	for (int i = 0; i < size; i++) {
		attr_RoiParam_read[i] =  w_val[i];
	}
	for (int i = size; i < 4; i++) {
		attr_RoiParam_read[i] = -1;
	}

	roi_dim = size;

}


//+------------------------------------------------------------------
/**
 * method: DataVampire::store_command_string
 * Store command string 
*/
//+------------------------------------------------------------------
void DataVampire::store_command_string(Tango::DevString w_val)
{
	std::string new_configuration(w_val);
	std::string dev_property("CommandString");
	
	Tango::Database *db = new Tango::Database(); 
	
	try {
		std::string dev_property("CommandString");
		Tango::DbDatum config(dev_property);
		Tango::DbData db_data_del, db_data_put;
		db_data_del.push_back(config);
		/* start data */
		config << new_configuration;
		/* end data */
		db_data_put.push_back(config);
		db->put_device_property(get_name(), db_data_put);
	}
	catch (CORBA::Exception &e) {
		delete db;	
		ERROR_STREAM << "write_CommandString(): failed to save parameters" << std::endl;		
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			__FUNCTION__);
	}	
	delete db;

	strncpy(*attr_CommandString_read, new_configuration.c_str(), 256);

	std::vector<std::string> results;
	string_explode(new_configuration, "/", &results);

	command_name = results[results.size()-1];
	size_t pos = new_configuration.find(command_name);
	if (pos != std::string::npos) { // if str2 found in str1
        	new_configuration.erase(pos-1, command_name.length()+1); // erase str2 from str1
    	}	
	device_name = new_configuration;

}



void DataVampire::extract_input_parameters(const Tango::DevVarLongArray *argin, int *mode, int *num_samples, int *bn_start, int *bn_end, int *roi)
{
	if (argin->length() < 2) {
		Tango::Except::throw_exception (
			(const char *)"Failed input parameters size",
			(const char *)"Minimum number is 4",
			__FUNCTION__);

	}	

	*mode = ((*argin)[0] % 10) & 0x1; // 20->0 21->1

	// return the last number
	if (*mode == 0) {
		*num_samples = (*argin)[1];
		*bn_start = 0;
		*bn_end = 0;		
	}
	else if (*mode == 1) {
		*bn_start = (*argin)[1];
		*bn_end = (*argin)[2];
		*num_samples = *bn_end - *bn_start + 1;
	}

	for (int i = 0; i < 4; i++) {
		roi[i] = attr_RoiParam_read[i];
	}
	

}




//+------------------------------------------------------------------
/**
 * method: DataVampire::free_cache_buffer
 * Free cache buffer 
*/
//+------------------------------------------------------------------
void DataVampire::free_cache_buffer(void)
{
	if (cache_buffer != NULL) {
		for (int i = 0; i < *attr_CacheSize_read; i++) {
			if (cache_buffer[i].databuf)
				free(cache_buffer[i].databuf);
			if (cache_buffer[i].datavalid)
				free(cache_buffer[i].datavalid);
			if (cache_buffer[i].sum)
				free(cache_buffer[i].sum);
			if (cache_buffer[i].mean)
				free(cache_buffer[i].mean);
			if (cache_buffer[i].horpos)
				free(cache_buffer[i].horpos);
			if (cache_buffer[i].verpos)
				free(cache_buffer[i].verpos);
			if (cache_buffer[i].horposstd)
				free(cache_buffer[i].horposstd);
			if (cache_buffer[i].verposstd)
				free(cache_buffer[i].verposstd);
			if (cache_buffer[i].validvalues)
				free(cache_buffer[i].validvalues);
		}
	}
	free(cache_buffer);

}


//+------------------------------------------------------------------
/**
 * method: DataVampire::allocate_cache_buffer
 * Allocate cache buffer 
*/
//+------------------------------------------------------------------
void DataVampire::allocate_cache_buffer(void)
{
	if (cache_buffer != NULL) {
		free_cache_buffer();
	}

	cache_buffer = (acq_buffer_t *) malloc(*attr_CacheSize_read * sizeof(acq_buffer_t));

	for (int i = 0; i < *attr_CacheSize_read; i++) {
		cache_buffer[i].databuf = 0;
		cache_buffer[i].datavalid = 0;
		cache_buffer[i].sum = 0;
		cache_buffer[i].mean = 0;	
		cache_buffer[i].horpos = 0;
		cache_buffer[i].verpos = 0;
		cache_buffer[i].horposstd = 0;	
		cache_buffer[i].verposstd = 0;
		cache_buffer[i].validvalues = 0;
		clock_gettime(CLOCK_REALTIME, &cache_buffer[i].timerec);
		cache_buffer[i].bn_start = 0;
		cache_buffer[i].bn_end = 0;		
		cache_buffer[i].num_samples = 0;
		cache_buffer[i].num_rows = 0;
		cache_buffer[i].num_cols = 0;
		cache_buffer[i].data_size = 0;		
	}	

	cache_idx = 0;

}


//+------------------------------------------------------------------
/**
 * method: DataVampire::store_data
 * Allocate cache buffer 
*/
//+------------------------------------------------------------------
void DataVampire::store_data(Tango::DeviceData *cmd_value, int32_t num_samples, int32_t bn_start, int32_t bn_end, int32_t *roi, struct acq_buffer_t **rec_data)
{
	std::vector<double> val;
	std::vector<std::string> val_string;

	if (*attr_EnableAcquisition_read == false) {
		Tango::Except::throw_exception (
			(const char *)"Failed to store data",
			(const char *)"Acquisition is disabled",
			__FUNCTION__);	
	}

	clock_gettime(CLOCK_REALTIME, &cache_buffer[cache_idx].timerec);
	cmd_extract_values(cmd_value,  val, val_string);

	int num_rows, num_cols;

	if (roi_dim == 2) {
		num_cols = roi[1] - roi[0] + 1;
		num_rows = 1;
	}
	else {
		num_rows = roi[3]-roi[1]+1,
		num_cols = roi[2]-roi[0]+1;
	}

	int32_t wave_size = val.size() / num_samples;

	if (((num_cols * num_rows * num_samples) != val.size()) && (*attr_PostAcqRoi_read == false)) {
		Tango::Except::throw_exception (
			(const char *)"Failed to store data",
			(const char *)"Data size does not match roi setting",
			__FUNCTION__);
	}

	if (((num_cols * num_rows * num_samples) > val.size()) && (*attr_PostAcqRoi_read == true)) {
		Tango::Except::throw_exception (
			(const char *)"Failed to store data",
			(const char *)"Data size too large",
			__FUNCTION__);
	}

	if ((*attr_PostAcqRoi_read == true) && (roi_dim == 4)) { 
		if (val.size() != (*attr_Width_read * *attr_Height_read * num_samples)) {
			Tango::Except::throw_exception (
				(const char *)"Failed to store data",
				(const char *)"Width/Height mismatch with image size",
				__FUNCTION__);		
		}
	
	}

	// reallocate buffer 
	if (val.size() != cache_buffer[cache_idx].data_size) {
		cache_buffer[cache_idx].databuf = (double *)realloc(cache_buffer[cache_idx].databuf, val.size() * sizeof(double));
		cache_buffer[cache_idx].datavalid = (bool *)realloc(cache_buffer[cache_idx].datavalid, val.size() * sizeof(bool));
		cache_buffer[cache_idx].data_size = val.size();
		cache_buffer[cache_idx].sum = (double *)realloc(cache_buffer[cache_idx].sum, num_samples * sizeof(double));
		cache_buffer[cache_idx].mean = (double *)realloc(cache_buffer[cache_idx].mean, num_samples * sizeof(double));
		cache_buffer[cache_idx].horpos = (double *)realloc(cache_buffer[cache_idx].horpos, num_samples * sizeof(double));
		cache_buffer[cache_idx].horposstd = (double *)realloc(cache_buffer[cache_idx].horposstd, num_samples * sizeof(double));
		if (num_rows > 1) {
			cache_buffer[cache_idx].verpos = (double *)realloc(cache_buffer[cache_idx].verpos, num_samples * sizeof(double));
			cache_buffer[cache_idx].verposstd = (double *)realloc(cache_buffer[cache_idx].verposstd, num_samples * sizeof(double));
		}
		cache_buffer[cache_idx].validvalues = (double *)realloc(cache_buffer[cache_idx].validvalues, num_samples * sizeof(double));
	}

	if (*attr_PostAcqRoi_read == false) {
		for (int i = 0; i < cache_buffer[cache_idx].data_size; i++) {
			cache_buffer[cache_idx].databuf[i] = val[i];
	
			if ((val[i] >= *attr_MinThreshold_read) && (val[i] <= *attr_MaxThreshold_read)) 
				cache_buffer[cache_idx].datavalid[i] = true;
			else
				cache_buffer[cache_idx].datavalid[i] = false;
		}
		
	}
	// ROI is not supported by command inout so get all the image stream with command inout
	else {
		int cnt = 0;
		
		for (int i = 0; i < num_samples; i++) {
			int offset = wave_size * i;
			//printf("num_samples = %d, wave_size = %d offset = %d val.size=%d\n",num_samples, wave_size, offset, val.size());
			double *data = &val[offset];
			// ImageWidth must be defined because command_inout does not report the Image width/length values
			if ((*attr_Width_read > 0) && (roi_dim == 4)) {
				for (int j = roi[1]; j <= roi[3]; j++) {
					for (int k = roi[0]; k <= roi[2]; k++) {
						cache_buffer[cache_idx].databuf[cnt] = data[j * *attr_Width_read + k];
						//cache_buffer[cache_idx].databuf[cnt] = k;
						if ((data[j * *attr_Width_read + k] >= *attr_MinThreshold_read) && 
							(data[j * *attr_Width_read + k] <= *attr_MaxThreshold_read)) 
							cache_buffer[cache_idx].datavalid[cnt] = true;
						else
							cache_buffer[cache_idx].datavalid[cnt] = false;


						cnt++;
					}
				}
			}
			else if (roi_dim == 2) {
				//printf("roi0 = %d roi1 = %d\n",roi[0],roi[1]);
				for (int k = roi[0]; k <= roi[1]; k++) {
					cache_buffer[cache_idx].databuf[cnt] = data[k];

					if ((data[k] >= *attr_MinThreshold_read) && (data[k] <= *attr_MaxThreshold_read)) 
						cache_buffer[cache_idx].datavalid[cnt] = true;
					else
						cache_buffer[cache_idx].datavalid[cnt] = false;
					cnt++;
				}	
			} 
		}



	}


	if (num_rows > 1) {
		for (int i = 0; i < num_rows*num_cols; i++) {
			attr_Image_read[i] = cache_buffer[cache_idx].databuf[i];
		}
		 num_rows_image = num_rows;
		 num_cols_image = num_cols;
	}
	
	
	/* performed in process_1D
	else if (num_rows == 1) {
		for (int i = 0; i < num_cols; i++) {
			attr_Spectrum_read[i] = cache_buffer[cache_idx].databuf[i];
		}
		num_cols_spectrum = num_cols;
	}*/

	cache_buffer[cache_idx].bn_start = bn_start;
	cache_buffer[cache_idx].bn_end = bn_end;		
	cache_buffer[cache_idx].num_samples = num_samples;
	if (roi_dim == 2) {
		cache_buffer[cache_idx].num_rows = 1;
		cache_buffer[cache_idx].num_cols = num_cols;		
	}	
	else if (roi_dim == 4) {
		cache_buffer[cache_idx].num_rows = num_rows;
		cache_buffer[cache_idx].num_cols = num_cols;
	}

	if (cache_buffer[cache_idx].num_rows == 1) {
		process_1D(cache_idx, *attr_AutoBackgroundSamples_read);
	}
	else if (cache_buffer[cache_idx].num_rows > 1) {
		process_2D(cache_idx, *attr_AutoBackgroundSamples_read);
	}

	*rec_data = &cache_buffer[cache_idx];	

	cache_idx++;
	if (cache_idx >= *attr_CacheSize_read)
		cache_idx = 0;

}


void DataVampire::process_2D(int32_t cache_idx, int32_t auto_bck_samples) 
{

	int32_t num_rows = cache_buffer[cache_idx].num_rows;
	int32_t num_cols = cache_buffer[cache_idx].num_cols;

	int32_t wave_size = num_rows * num_cols;
	int32_t data_size = cache_buffer[cache_idx].data_size;
	int32_t num_samples = cache_buffer[cache_idx].num_samples;

	double *hor_profile_bck, *ver_profile_bck;
	uint32_t *hor_profile_bck_cnt, *ver_profile_bck_cnt;

	// for Image display
	num_rows_image = num_rows;
	num_cols_image = num_cols;

	hor_profile_bck = (double *) calloc(num_cols, sizeof(double));
	ver_profile_bck = (double *) calloc(num_rows, sizeof(double));
	hor_profile_bck_cnt = (uint32_t *) calloc(num_cols, sizeof(uint32_t));
	ver_profile_bck_cnt = (uint32_t *) calloc(num_rows, sizeof(uint32_t));

	for (int i = 0; i < num_samples; i++) {

		memset(hor_profile_bck, 0, sizeof(float) * num_cols);
		memset(ver_profile_bck, 0, sizeof(float) * num_rows);
		memset(hor_profile_bck_cnt, 0, sizeof(uint32_t) * num_cols);
		memset(ver_profile_bck_cnt, 0, sizeof(uint32_t) * num_rows);

		if (auto_bck_samples) {
			// calculate horizontal background profile from top and bottom auto_bck_samples rows
			for (int j = 0; j < *attr_AutoBackgroundSamples_read; j++) {			
				for (int k = 0; k < num_cols; k++) {
					int idx = j*num_cols+k+wave_size*i;
					if (cache_buffer[cache_idx].datavalid[idx]) {
						hor_profile_bck[k] += cache_buffer[cache_idx].databuf[idx];	
						hor_profile_bck_cnt[k]++;	
					}
					idx = (num_rows-j-1)*num_cols+k+wave_size*i;
					if (cache_buffer[cache_idx].datavalid[idx]) {
						hor_profile_bck[k] += cache_buffer[cache_idx].databuf[idx];	
						hor_profile_bck_cnt[k]++;	
					}
				}
			}

			// calculate vertical background profile from left and right auto_bck_samples columns
			for (int j = 0; j < *attr_AutoBackgroundSamples_read; j++) {
				// rows				
				for (int k = 0; k < num_rows; k++) {
					int idx = k*num_cols+j+wave_size*i;
					if (cache_buffer[cache_idx].datavalid[idx]) {
						ver_profile_bck[k] += cache_buffer[cache_idx].databuf[idx];
						//printf("k=%d ver_bck=%f idx=%d \n",k,cache_buffer[cache_idx].databuf[idx],idx);
						ver_profile_bck_cnt[k]++;	
					}
					//int idx2 = (num_rows-k-1)*num_cols-j+wave_size*i;
					int idx2 = idx + (num_cols - 1) - j;
					if (cache_buffer[cache_idx].datavalid[idx2]) {
						ver_profile_bck[k] += cache_buffer[cache_idx].databuf[idx2];	
						//printf("k=%d ver_bck=%f idx2=%d \n",k,cache_buffer[cache_idx].databuf[idx2],idx2);
						ver_profile_bck_cnt[k]++;	
					}
					//printf("k=%d ver_bck=%f ver_bck_cnt=%d idx=%d idx2=%d\n",k,ver_profile_bck[k],ver_profile_bck_cnt[k],idx,idx2);
				}
			}
	

			for (int j = 0; j < num_cols; j++) {
				if (hor_profile_bck_cnt[j] > 0) {
					hor_profile_bck[j] = hor_profile_bck[j] / (double)hor_profile_bck_cnt[j];
				}
				else
					hor_profile_bck[j] = 0;
			}
			for (int j = 0; j < num_rows; j++) {
				if (ver_profile_bck_cnt[j] > 0) {
					ver_profile_bck[j] = ver_profile_bck[j] / (double)ver_profile_bck_cnt[j];
				}
				else
					ver_profile_bck[j] = 0;
			}

			if (i == 0) {
				for (int j = 0; j < num_cols; j++) 
					attr_HorBackground_read[j] = hor_profile_bck[j];
				for (int j = 0; j < num_rows; j++)
					attr_VerBackground_read[j] = ver_profile_bck[j];
			}

			// remove horizontal backgroud
			for (int j = 0; j < num_rows; j++) {
				for (int k = 0; k < num_cols; k++) {
					int idx = j * num_cols + k + wave_size*i;
					cache_buffer[cache_idx].databuf[idx] -= hor_profile_bck[k];
				}
			}

			// remove vertical backgroud
			for (int j = 0; j < num_cols; j++) {
				for (int k = 0; k < num_rows; k++) {
					int idx = k * num_cols + j + wave_size*i;
					cache_buffer[cache_idx].databuf[idx] -= ver_profile_bck[k];
				}
			}


			for (int j = 0; j < num_rows; j++) {
				for (int k = 0; k < num_cols; k++) {
					int idx = j * num_cols + k + wave_size*i;
					if (cache_buffer[cache_idx].databuf[idx] < 0)
						cache_buffer[cache_idx].databuf[idx] = 0;						
				}
			}
			
		}  // auto_bck_samples


		// calculate position, intensity and rms of the image
		double hor_pos = 0, ver_pos = 0, mean = 0;
		double std_hor_pos = 0, std_ver_pos = 0;
		double total_sum = 0;
		int32_t valid_values = 0;

		for(int j = 0; j < num_rows; j++) {
			for(int k = 0; k < num_cols; k++) {
				int idx = j * num_cols + k + wave_size*i;

				if (i == 0)
					attr_Image_read[idx] = cache_buffer[cache_idx].databuf[idx];

				if (cache_buffer[cache_idx].datavalid[idx]) {
					hor_pos += (cache_buffer[cache_idx].databuf[idx] * (k+1));
					ver_pos += (cache_buffer[cache_idx].databuf[idx] * (j+1));
					total_sum += cache_buffer[cache_idx].databuf[idx];
					valid_values++;
				}
			}
		}

		hor_pos /= total_sum;
		ver_pos /= total_sum;

		for(int i = 0; i < num_rows; i++) {
			for(int j = 0; j < num_cols; j++) {
				std_hor_pos += (i-hor_pos)*(i-hor_pos);
				std_ver_pos += (j-ver_pos)*(j-ver_pos);
			}
		}

		if (total_sum > 0) {
			std_hor_pos = pow(std_hor_pos / total_sum, 0.5);     
			std_ver_pos = pow(std_ver_pos / total_sum, 0.5);   
		}

		cache_buffer[cache_idx].sum[i] = total_sum;
		
		cache_buffer[cache_idx].mean[i] =  total_sum / (double)valid_values;
		cache_buffer[cache_idx].horpos[i] = hor_pos;
		cache_buffer[cache_idx].verpos[i] = ver_pos;		
		cache_buffer[cache_idx].horposstd[i] = std_hor_pos;
		cache_buffer[cache_idx].verposstd[i] = std_ver_pos;	
		cache_buffer[cache_idx].validvalues[i] = valid_values;

		if (i == 0) {
			*attr_Sum_read =  cache_buffer[cache_idx].sum[i];
			*attr_Mean_read =  cache_buffer[cache_idx].mean[i];
			*attr_HorPos_read =  cache_buffer[cache_idx].horpos[i];			
			*attr_VerPos_read =  cache_buffer[cache_idx].verpos[i];		
			*attr_HorPosStd_read =  cache_buffer[cache_idx].horposstd[i];
			*attr_VerPosStd_read =  cache_buffer[cache_idx].verposstd[i];	
			*attr_ValidValues_read =  cache_buffer[cache_idx].validvalues[i];					
		}


	}
	
	free(hor_profile_bck);
	free(ver_profile_bck);
	free(hor_profile_bck_cnt);
	free(ver_profile_bck_cnt);	

}


void DataVampire::process_1D(int32_t cache_idx, int32_t auto_bck_samples) 
{
	int32_t wave_size = cache_buffer[cache_idx].num_rows *
		cache_buffer[cache_idx].num_cols;
	int32_t num_samples = cache_buffer[cache_idx].num_samples;
	int32_t data_size = cache_buffer[cache_idx].data_size;

	//printf("GG 1 process_1D cache_idx=%d num_cols=%d\n",cache_idx,cache_buffer[cache_idx].num_cols);

	// for spectrum display
	num_cols_spectrum = cache_buffer[cache_idx].num_cols;

	for (int i = 0; i < num_samples; i++) {
		double sum = 0;
		double mean = 0;
		double sum_left = 0;
		double sum_right = 0;
		double sum_top = 0;
		double sum_bottom = 0;
		double offset = 0;
		double slope = 0;
		double pos = 0;
		double std = 0;
		int cnt_left = 0, cnt_right = 0, cnt = 0;
		double valid_values = 0;
		if (auto_bck_samples) {
			for (int j = 0; j < auto_bck_samples; j++) {
				if (cache_buffer[cache_idx].datavalid[i*wave_size+j]) {									
					double val = cache_buffer[cache_idx].databuf[i*wave_size+j];
					sum_left += val;
					cnt_left++;
				}
				if (cache_buffer[cache_idx].datavalid[i*wave_size+j]) {									
					double val = cache_buffer[cache_idx].databuf[i*wave_size+(wave_size-auto_bck_samples)+j];
					sum_right += val;
					cnt_right++;
				}				
			}
			if (cnt_left)
				sum_left /= (double)cnt_left;
			else
				sum_left = 0;
			if (cnt_right)
				sum_right /= (double)cnt_right;
			else
				sum_right = 0;
				
			if (auto_bck_samples > 0) {			
				slope = (sum_right - sum_left) / 
					((double) (wave_size - (double)(cnt_left+cnt_right)/2));
				offset = sum_left;
				if (offset > sum_right)
					offset = sum_right;
			}
			else { 
				slope = 0; offset = 0;
			}
		}

	
		


		// get sum, mean and position
		for (int j = 0; j < cache_buffer[cache_idx].num_cols; j++) {
			//printf("j=%d idx=%d\n",j,i*wave_size+j);
			double val = (cache_buffer[cache_idx].databuf[i*wave_size+j] + slope * (double)(cache_buffer[cache_idx].num_cols - j) - offset);

			// for spectrum displaying
			if (i == 0)
				attr_Spectrum_read[j] = val; 

			if (cache_buffer[cache_idx].datavalid[i*wave_size+j]) {
				sum += val; 
				pos += val * (double)(j + 1);
				cnt++;
			}				
		}
		mean = sum / (double) cnt;
		pos /= sum;


		// get std
		cnt = 0;
		for (int j = 0; j < cache_buffer[cache_idx].num_cols; j++) {
			double val = (cache_buffer[cache_idx].databuf[i*wave_size+j] + slope * (double)j - offset);
			if (cache_buffer[cache_idx].datavalid[i*wave_size+j]) {
				std += ((double)j - pos) * ((double)j - pos); cnt++;
			}				
		}		
		if (cnt != 0)
			std = pow(std / (double) cnt, 0.5);
		else
			std = 0;

		cache_buffer[cache_idx].sum[i] = sum;
		cache_buffer[cache_idx].mean[i] = mean;
		cache_buffer[cache_idx].horpos[i] = pos;
		cache_buffer[cache_idx].horposstd[i] = std;
		cache_buffer[cache_idx].validvalues[i] = cnt;


	}

}



//+------------------------------------------------------------------
/**
 * method: DataVampire::recover_data
 * Recover data
 */
//+------------------------------------------------------------------
void DataVampire::recover_data(const Tango::DevVarLongArray *argin, struct acq_buffer_t **rec_data)
{

	std::vector<double> cmd_input_value;
	std::vector<std::string> cmd_input_value_str;
	std::vector<double> cmd_output_value;
	std::vector<std::string> cmd_output_value_str;

	struct acq_buffer_t *tmp;

	int32_t mode, num_samples, bn_start, bn_end, roi[4];

	for (unsigned int i = 0; i < argin->length(); i++) {
		cmd_input_value.push_back((*argin)[i]);
	}	

	if (*attr_PostAcqRoi_read == false) {
		for (unsigned int i = 0; i < roi_dim; i++) {
			cmd_input_value.push_back(attr_RoiParam_read[i]);
		}	
	}
	
	Tango::DeviceData Din = cmd_insert_values(cmd_in_type, cmd_input_value, cmd_input_value_str);

	extract_input_parameters(argin, &mode, &num_samples, &bn_start, &bn_end, roi);	

	Tango::DeviceData Dout = device->command_inout(command_name, Din);

	store_data(&Dout, num_samples, bn_start, bn_end, roi, rec_data);

}

// //--------------------------------------------------------
// /**
//  *	Read attribute VerPosStd related method
//  *	Description: 
//  *
//  *	Data type:	Tango::DevDouble
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void DataVampire::read_VerPosStd(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "DataVampire::read_VerPosStd(Tango::Attribute &attr) entering... " << endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_VerPosStd_read);
// 	/* clang-format off */
// }

// //--------------------------------------------------------
// /**
//  *	Read attribute VerPosStd related method
//  *	Description: 
//  *
//  *	Data type:	Tango::DevDouble
//  *	Attr type:	Scalar
//  */
// //--------------------------------------------------------
// void DataVampire::read_VerPosStd(Tango::Attribute &attr)
// {
// 	DEBUG_STREAM << "DataVampire::read_VerPosStd(Tango::Attribute &attr) entering... " << endl;
// 	/* clang-format on */
// 	//	Set the attribute value
// 	attr.set_value(attr_VerPosStd_read);
// 	/* clang-format off */
// }


/*----- PROTECTED REGION END -----*/	//	DataVampire::namespace_ending
} //	namespace
